@@ -25,11 +25,115 @@ except ImportError:
     import gnome.canvas
     canvasmodule = gnome.canvas
 
+import runtime
+if runtime.use_cairo_widgets:
+    import math
+
 import mpd
 import soundcard
 import cfg
 import runtime
 
+class CairoPianoKeyboard(gtk.DrawingArea):
+    def __init__(self, num_octaves, lowest_c, key_w=15):
+        gtk.DrawingArea.__init__(self)
+        self.connect("expose_event", self.expose)
+        self.connect("button-press-event", self._on_button_press)
+        self.set_events(gtk.gdk.BUTTON_PRESS_MASK | gtk.gdk.POINTER_MOTION_MASK | gtk.gdk.ENTER_NOTIFY_MASK | gtk.gdk.LEAVE_NOTIFY_MASK)
+        # Piano stuff
+        self.m_num_octaves = num_octaves
+        self.m_lowest_c = mpd.MusicalPitch.new_from_notename(lowest_c)
+        self.m_white_h = key_w * 3.4
+        self.m_black_h = key_w * 2.0
+        self.m_key_w = key_w
+        self.m_last_redraw = 0
+        self.m_black_w = 0.6
+        self.set_size_request(int(num_octaves * 7 * key_w + 1), int(self.m_white_h + 1))
+        self.m_marks = {}
+    def _redraw(self):
+        self.window.invalidate_rect(gtk.gdk.Rectangle(0, 0,
+            self.allocation.width, self.allocation.height), True)
+        return False
+    def on_event(self, widget, event):
+        if event.type == gtk.gdk.MOTION_NOTIFY:
+            self.on_motion_notify(widget, event)
+    def _on_button_press(self, drawingarea, event):
+        assert event.x >= 0
+        if event.x < self.m_pos_x or event.x > self.m_num_octaves * 7 * self.m_key_w + self.m_pos_x:
+            # Click outside the keys are ignored.
+            return
+        click_x = event.x - self.m_pos_x
+        x = click_x / self.m_key_w
+        octave_w = 7 * self.m_key_w
+        clicked_octave = int(click_x / octave_w)
+        clicked_white = int((click_x % octave_w) / self.m_key_w)
+        black_clicked = 0
+        if event.y < self.m_black_h:
+            if clicked_white in (0, 1, 3, 4, 5):
+                # How many pixels into the white key does the black
+                # key start:
+                bpos1 = self.m_key_w - self.m_key_w * self.m_black_w / 2
+                if click_x % self.m_key_w >= bpos1:
+                    black_clicked = 1
+            if clicked_white in (1, 2, 4, 5, 6):
+                # How many pixels into the white key does the black
+                # key end:
+                bpos2 = self.m_key_w * self.m_black_w / 2
+                if click_x % self.m_key_w <= bpos2:
+                    black_clicked = -1
+        clicked_on = self.m_lowest_c.clone()
+        clicked_on.m_notename_i = clicked_white
+        clicked_on.m_accidental_i = black_clicked
+        clicked_on.m_octave_i += clicked_octave
+        self.on_button_press_event(None, event, clicked_on.semitone_pitch())
+    def expose(self, widget, event):
+        self.context = widget.window.cairo_create()
+        # set a clip region for the expose event
+        self.context.rectangle(event.area.x, event.area.y,
+                               event.area.width, event.area.height)
+        self.context.clip()
+        self.draw(self.context)
+        return False
+    def draw(self, context):
+        num_key = 7 * self.m_num_octaves
+        width = num_key * self.m_key_w
+        self.m_pos_x = pos_x = (self.allocation.width - width) / 2 + 0.5
+        pos_y = .5
+        context.set_line_width(1.2)
+        context.save()
+        for n in range(num_key):
+            context.rectangle(pos_x + n * self.m_key_w, pos_y,
+                              self.m_key_w, self.m_white_h)
+        context.set_source_rgb(1, 1, 1)
+        context.fill_preserve()
+        context.set_source_rgb(0, 0, 0)
+        context.stroke()
+        context.restore()
+
+        # Draw black keys
+        context.save()
+        for oct in range(self.m_num_octaves):
+            for n in (0, 1, 3, 4, 5):
+                context.rectangle(pos_x + oct*7*self.m_key_w + (n+1-self.m_black_w/2) * self.m_key_w,
+                            pos_y,
+                            self.m_key_w * self.m_black_w, self.m_black_h)
+                context.set_source_rgb(0, 0, 0)
+                context.fill_preserve()
+        context.stroke()
+        context.restore()
+        for acc, step in self.m_marks.values():
+            if acc == 0:
+                context.arc(pos_x + (step + 0.5) * self.m_key_w, self.m_white_h-(self.m_white_h - self.m_black_h) / 2,
+                            self.m_key_w * 0.35, 0, 2 * math.pi)
+            else:
+                context.arc(pos_x + (step + 0.5 + 0.5 * acc) * self.m_key_w, self.m_black_h * 0.5,
+                            self.m_key_w * 0.25, 0, 2 * math.pi)
+            context.set_source_rgb(0, 0.8, 0)
+            context.fill_preserve()
+            context.stroke()
+    def grab_focus_first_sensitive_button(self):
+        pass
+
 class PianoKeyboard(canvasmodule.Canvas):
     def __init__(self, lowest_tone, highest_tone, size):
         """
@@ -81,7 +185,7 @@ class PianoKeyboard(canvasmodule.Canvas):
         if event.type == gtk.gdk.BUTTON_PRESS:
             self.on_button_press_event(group, event, midi_int)
 
-class PianoOctaveWithAccelName(PianoKeyboard):
+class CanvasPianoOctaveWithAccelName(PianoKeyboard):
     def __init__(self, callback, keys):
         PianoKeyboard.__init__(self, 'c', 'b', (30, 100, 60))
         self.m_callback = callback
@@ -111,17 +215,64 @@ class PianoOctaveWithAccelName(PianoKeyboard):
         elif event.button == 1:
             self.m_callback(mpd.int_to_notename(midi_int))
 
+class CairoPianoOctaveWithAccelName(CairoPianoKeyboard):
+    def __init__(self, callback, keys):
+        CairoPianoKeyboard.__init__(self, 1, "c", 40)
+        self.m_callback = callback
+        self.m_keys = keys
+    def draw(self, context):
+        CairoPianoKeyboard.draw(self, context)
+        context.save()
+        context.select_font_face("Sans")
+        context.set_font_size(24)
+        text_h = context.text_extents('A')[3]
+        for idx, n in enumerate((0, 2, 4, 5, 7, 9, 11)):
+            context.new_path()
+            text_w = context.text_extents(self.m_keys[n])[4]
+            context.move_to(self.m_pos_x + idx*self.m_key_w + self.m_key_w / 2 - text_w / 2, self.m_black_h + (self.m_white_h - self.m_black_h) / 2 + text_h / 2)
+            context.text_path(self.m_keys[n])
+            context.fill()
+            context.stroke()
+        context.restore()
+        context.save()
+        context.select_font_face("Sans")
+        context.set_font_size(24)
+        for idx, n in enumerate((1, 3, None, 6, 8, 10)):
+            if n:
+                context.new_path()
+                text_w = context.text_extents(self.m_keys[n])[4]
+                context.move_to(self.m_pos_x + (idx + 1)*self.m_key_w - text_w/2, self.m_black_h * 0.8)
+                context.text_path(self.m_keys[n])
+                context.set_source_rgb(1, 1, 1)
+                context.fill()
+                context.stroke()
+        context.restore()
+    def on_button_press_event(self, group, event, midi_int):
+        if event.button == 3:
+            soundcard.play_note(cfg.get_int('config/preferred_instrument'),
+                           4, 0, midi_int,
+                           cfg.get_int('config/preferred_instrument_velocity'))
+        elif event.button == 1:
+            self.m_callback(mpd.int_to_notename(midi_int))
 
-class IntervalPianoWidget(PianoKeyboard):
+if runtime.use_cairo_widgets:
+    PianoClass = CairoPianoKeyboard
+    PianoOctaveWithAccelName = CairoPianoOctaveWithAccelName
+else:
+    PianoClass = PianoKeyboard
+    PianoOctaveWithAccelName = CanvasPianoOctaveWithAccelName
+
+class BaseIntervalPianoWidget(PianoClass):
     def __init__(self, callback):
-        PianoKeyboard.__init__(self, "c,", "b''", (15, 50, 30))
+        if runtime.use_cairo_widgets:
+            CairoPianoKeyboard.__init__(self, 4, "c,", 18)
+        else:
+            PianoKeyboard.__init__(self, "c,", "b''", (15, 50, 30))
         self.m_clicked_tones = []
         self.m_callback = callback
         self.m_marks = {}
     def set_first_note(self, tone):
-        for key in self.m_marks.keys():
-            self.m_marks[key].destroy()
-            del self.m_marks[key]
+        self.clear()
         self.m_clicked_tones = [tone.semitone_pitch()]
         self.mark(self.m_clicked_tones[-1])
     def on_button_press_event(self, group, event, midi_int):
@@ -178,6 +329,25 @@ class IntervalPianoWidget(PianoKeyboard):
             assert i in self.m_blacks
             return self.m_blacks.index(i), 0
 
+class CairoIntervalPianoWidget(BaseIntervalPianoWidget):
+    def __init__(self, callback):
+        BaseIntervalPianoWidget.__init__(self, callback)
+    def mark(self, midi_int):
+        n = mpd.MusicalPitch.new_from_int(midi_int)
+        assert not self.m_marks
+        self.m_marks[n.get_notename()] = (n.m_accidental_i, n.steps() - self.m_lowest_c.steps())
+        self._redraw()
+    def unmark(self, midi_int):
+        for n in self.m_marks.keys():
+            if mpd.notename_to_int(n) == midi_int:
+                del self.m_marks[n]
+    def clear(self):
+        self.m_marks = {}
+
+if runtime.use_cairo_widgets:
+    IntervalPianoWidget = CairoIntervalPianoWidget
+else:
+    IntervalPianoWidget = BaseIntervalPianoWidget
 
 class IntervalButtonsWidget(gtk.Table, cfg.ConfigUtils):
     def __init__(self, exname, name, callback, sensicallback, vars_to_watch):
@@ -564,13 +734,14 @@ class AccordionWidget(canvasmodule.Canvas):
     def know_directions(self):
         return 1
 
+inputwidget_names = [_("Buttons")]
+if runtime.has_gnome_canvas() or runtime.use_cairo_widgets:
+    inputwidget_names.append(_("Piano"))
 if runtime.has_gnome_canvas():
-    inputwidget_names = [_("Buttons"), _("Piano"), _("Guitar"), _("Bass"),
+    inputwidget_names.extend([_("Guitar"), _("Bass"),
                  _("5 string bass"), _("6 string bass"),
                  _("Accordion B griff"), _("Accordion C griff"),
-                 _("Accordion (system used in Finland)")]
-else:
-    inputwidget_names = [_("Buttons")]
+                 _("Accordion (system used in Finland)")])
 
 def name_to_inputwidget(name, callback):
     if name == _("Piano"):
