@@ -1,5 +1,5 @@
 # GNU Solfege - ear training for GNOME
-# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005  Tom Cato Amundsen
+# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006  Tom Cato Amundsen
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -235,15 +235,21 @@ class Gui(abstract.IntervalGui):
         elif ret == Teacher.ERR_PICKY:
             self.g_flashbar.flash(_("You have to solve this question first."))
     def on_start_practise(self):
+        # First, we have to empty the cfg database because we will
+        # copy the values from the lesson header.
+        for i in range(self.get_int('maximum_number_of_intervals')):
+            self.set_list('ask_for_intervals_%i' % i, [])
         # If ask_for_intervals_0 is not set, then we run in custom_mode
         # where the user configure the exercise on her own.
         if 'ask_for_intervals_0' in self.m_t.m_P.header:
-            self.set_int('number_of_intervals', self.m_t.m_P.header.get('number_of_intervals', 1))
             self.set_bool('disable_unused_intervals', self.m_t.m_P.header.disable_unused_intervals)
             for i in range(self.get_int('maximum_number_of_intervals')):
                 if 'ask_for_intervals_%i' % i in self.m_t.m_P.header:
                     self.set_list('ask_for_intervals_%i' % i,
                       self.m_t.m_P.header['ask_for_intervals_%i' % i])
+                else:
+                    break
+            self.set_int('number_of_intervals', i)
         if self.m_t.m_P.header.lesson_heading:
             self.set_lesson_heading(self.m_t.m_P.header.lesson_heading)
         else:
